// -----------------------------
// ------- CUDA Includes -------
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

// -----------------------------
// -------- C++ Includes -------
#include <stdio.h>
#include <iostream>
#include <string>

// -------------------------------
// ------- Matlab Includes -------
#include <matrix.h>
#include <mex.h>

#define HANDLE_CUDA( err )  (Handle_Error( err, __FILE__, __LINE__ ))

/*********************************************************************************************
 * Note: Can add to Build Event to automatically Copy Elsewhere
 *  ECHO F | XCOPY /Y "D:\CudaFEM\x64\Debug\CudaFEM.mexw64" "D:\CudaFEM\Matlab\CudaFEM.mexw64"
 *********************************************************************************************/

/******************************************************
 *@brief Error Handling Function of all the cuda Calls
 ****************************************************/
void Handle_Error(cudaError_t err, const char *file, int line) {
	if (err != cudaSuccess) {
		std::string Message = std::string(cudaGetErrorString(err)) + std::string(" in ") + std::string(file) + std::string(" at line: ") + std::to_string(line) + std::string("\n");
		mexErrMsgTxt(Message.c_str());
	}
}

/***************************************************************
 *@brief Transforms a Colum-Major Order Array to Row-Major Order
 **************************************************************/
template <typename Type>
void ColToRow(Type *input, Type *output, int Nx, int Ny) {
	int counter = 0;
	for (int i = 0; i < Ny; i++) {
		for (int j = 0; j < Nx; j++) {
			int index = j * Ny + i;
			output[counter] = input[index];
			counter++;
		}
	}
}

/****************************************************************
 *@brief Transforms a Row-Major Order Array to Column-Major Order
 ****************************************************************/
template <typename Type>
void RowToCol(Type *output, Type* input, int Nx, int Ny) {
	int counter = 0;
	for (int i = 0; i < Ny; i++) {
		for (int j = 0; j < Nx; j++) {
			int index = j * Ny + i;
			output[index] = input[counter];
			counter++;
		}
	}
} 

/*******************************************************************
 *@brief Finds maximum of an array and stores it in the output Array
 ******************************************************************/
__global__ void ReduceMax(float *In,float *Out,int N){
	unsigned int tx = threadIdx.x;
	unsigned int bx = blockIdx.x + gridDim.x * blockIdx.y;
	unsigned int tix = threadIdx.x + blockIdx.x * blockDim.x;
	unsigned int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	unsigned int tid = tix + tiy * blockDim.x * gridDim.x;

	// DEFINE SAHRED MEMORY
	__shared__ float Local[1024];
	Local[tx] = 0.0f;


	if(tid < N){
		Local[tx] = In[tid];
		__syncthreads();

		//Reduction
		int Step = 512;
		for (int i = 0; i < 9; i++){
			if (tx < Step){
				if(Local[tx + Step] > Local[tx]){
					Local[tx] = Local[tx + Step];
				}
			}
			// SYCN FOR NEXT STAGE
			__syncthreads();
			Step = Step / 2;
		}

		// SAVE TO GLOBAL MEMORY
		if(tx == 0){Out[bx] = Local[0];}

	} 
}

/*************************************************************************************
 *@brief Main Computing Routine for the FEM Method
 *@param RAir: Error of Aired Condition
 *@param VAir: Potential of the Aired Condition
 *@param Omega: SuperRelaxation Coefficient
 *@param Mask : mask of the StripLines
 *@param Nx - Amount of Elements in the X-Dimension
 *@param Ny - Amount of Elements in the Y-Dimension
 *@param Grid8 - Use grid of 8 Points instead of 4 Points.
 * NOTE: This kernel could be eventualy improved by using shared memory
 *  THis will eliminate lots of Memory Accesses at a cost of extra indexing complexity
 *************************************************************************************/
__global__ void CalculatePotencials( float *RAir_Out, float *VAir_In,float *VAir_Out,
									 int *Mask, int Nx, int Ny,float Omega,bool Grid8) {

	unsigned int tix = threadIdx.x + blockIdx.x * blockDim.x;
	unsigned int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	if ((tix < (Nx - 1)) && (tiy < (Ny - 1)) && (tix > 0) && (tiy > 0)) {
		// Mesh Indexing
		unsigned int Center = tix + tiy * Nx;
		if (Mask[Center] != 1) {
			unsigned int IL = tix - 1 + tiy * Nx;      // Left Index
			unsigned int IR = tix + 1 + tiy * Nx;      // Right Index
			unsigned int IT = tix + (tiy - 1)*Nx;      // Top Index
			unsigned int IB = tix + (tiy + 1)*Nx;      // Bottom Index
			unsigned int IUR = tix + 1 + (tiy - 1)*Nx; // UpperRight Index
			unsigned int IUL = tix - 1 + (tiy - 1)*Nx; // UpplerLeft Index
			unsigned int ILR = tix + 1 + (tiy + 1)*Nx; // LowerRight Index
			unsigned int ILL = tix - 1 + (tiy + 1)*Nx; // LowerLeft Index

			// -------------------------------------------------
			// ------ Get nearby field values from global ------
			float VT = VAir_In[IT];
			float VB = VAir_In[IB];
			float VL = VAir_In[IL];
			float VR = VAir_In[IR];
			float VUR = VAir_In[IUR];
			float VUL = VAir_In[IUL];
			float VLR = VAir_In[ILR];
			float VLL = VAir_In[ILL]; 
			float WC = VAir_In[Center];

			// Compute Weights
			float C0 = 1.0f / sqrtf(2); // 0.7071 Weight
			float C1 = 1.0f - C0;       // 0.2929 Weight

			float Rep0 = 1.0f/4.0f * (VT + VB + VL + VR);
			float Rep1 = 1.0f/4.0f * (VUR + VUL + VLL + VLR);

			// Compute Residuum
			float Rep8 = C0 * Rep0 + C1 * Rep1 - WC;
			float Rep4 = Rep0 - WC;


			if(Grid8){
				// ---------------------------------------
				// ------- Calculate from 8 Points -------
				VAir_Out[Center] = WC + Omega * Rep8;
				RAir_Out[Center] = Rep8;
			}else{
				// ---------------------------------------
				// ------- Calculate from 4 Points -------
				VAir_Out[Center] = WC + Omega * Rep4;
				RAir_Out[Center] = Rep4;
			}

		}
	}
}


/**************************************************
 *@brief Main Matlab Function
 *@param nlhs - Amount of LeftHandSide Parameters
 *@param nrhs - Amount of RightHandside Parameters
 *@param plhs - Array of Output Arguments
 *@param prhs - Array of Input Arguments 
 ***************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){

	const mxArray *M_In_VAir  = prhs[0];
	const mxArray *M_In_RAir  = prhs[1];
	const mxArray *M_In_Mask  = prhs[2];
	const mxArray *M_In_Args  = prhs[3];
	
	// ----------------------------------------
	// ------- Check Amount of Aguments -------
	if (nrhs < 4) { mexErrMsgTxt("Not Enough Input Arguments!\n");}
	if (nlhs < 3) { mexErrMsgTxt("Not Enough Output Arguments!\n"); }

	// ----------------------------------------------
	// ------- Check Complexity of Input Data -------
	for(int i = 0;i < 4;i++){
		if(mxIsComplex(prhs[i])){
			std::string str = std::string("Error Argument") + std::to_string(i) + std::string(" cannot be complex!\n");
			mexErrMsgTxt(str.c_str());
		}
	}

	// ------------------------------
	// ------- CHECK CLASS ID -------
	if (mxGetClassID(M_In_VAir) != mxSINGLE_CLASS) { mexErrMsgTxt("VAir: Error Expected Single Precision!\n"); }
	if (mxGetClassID(M_In_RAir) != mxSINGLE_CLASS) { mexErrMsgTxt("RAir: Error Expected Single Precision!\n"); }
	if (mxGetClassID(M_In_Mask) != mxINT32_CLASS)  { mexErrMsgTxt("Mask: Error Expected In32 Precision!\n");   }

	// -------------------------------------------
	// ------- Get Pointers to Data Arrays -------
	float *In_VAir = static_cast<float*>(mxGetData(M_In_VAir)); 
	float *In_RAir = static_cast<float*>(mxGetData(M_In_RAir)); 
	float *In_Args = static_cast<float*>(mxGetData(M_In_Args)); 
	int   *In_Mask = static_cast<int*>(mxGetData(M_In_Mask)); 

	float *VAir = nullptr;
	float *RAir = nullptr;
	int   *Mask = nullptr;

	// --------------------------------------
	// ------- Check Input Dimensions -------
	bool XDimCheck = false;
	bool YDimCheck = false;
	const mwSize *DimsAir = mxGetDimensions(M_In_VAir);
	const mwSize *DimsRAir = mxGetDimensions(M_In_RAir);
	const mwSize *DimsMask = mxGetDimensions(M_In_Mask);
	if((DimsAir[0] == DimsRAir[0]) && (DimsRAir[0] ==  DimsMask[0])){YDimCheck = true;}
	if((DimsAir[1] == DimsRAir[1]) && (DimsRAir[1] ==  DimsMask[1])){XDimCheck = true;}
	if(!XDimCheck){mexErrMsgTxt("Error Invalid Input Dimensions!\n");}
	if(!YDimCheck){mexErrMsgTxt("Error Invalid Input Dimensions!\n");}
	

	// Figure Out Dimensions and Computing Params
	if (mxGetNumberOfElements(M_In_Args) < 5) {
		std::string Message = std::string("Error Expects these arguments: ");
		Message += std::string("Arg[0]: SuperRelaxation Coefficient\n");
		Message += std::string("Arg[1]: Maximum Amount of Iterations\n");
		Message += std::string("Arg[2]: Convergence Threshold\n");
		Message += std::string("Arg[3]: Memory Synchronization (0/1)\n");
		Message += std::string("Arg[4]: Grid8(1) / Grid4(0)\n");
		mexErrMsgTxt(Message.c_str());
	}

	// -------------------------------
	// ------- Parse Parameters ------
	int Nx = static_cast<int>(DimsAir[1]);
	int Ny = static_cast<int>(DimsAir[0]);
	float Omega		= In_Args[0];
	int MaxIter		= static_cast<int>(In_Args[1]);
	float Thresh	= static_cast<float>(In_Args[2]);
	int SwitchInOut = static_cast<int>(In_Args[3]);
	int UseGrid8	= static_cast<int>(In_Args[4]);
	bool SwitchDir	= false;
	bool Grid8		= false;
	if(SwitchInOut > 0){SwitchDir = true;}
	if(UseGrid8 > 0){Grid8 = true;}

#ifdef DEBUG_PRINT
	mexPrintf("Dimensions| X: %d Y:%d\n", Nx, Ny);
	mexPrintf("Omega: %4.3f Thresh: %8.7f MaxIter:%d \n", Omega,Thresh,MaxIter);
#endif

	float *Dev_RAir   = nullptr;
	float *Dev_VAir_0 = nullptr;
	float *Dev_VAir_1 = nullptr;
	int   *Dev_Mask   = nullptr;
	float *Dev_Max_0  = nullptr;
	float *Dev_Max_1  = nullptr;
	float *Dev_Max_2  = nullptr;


	int RuntimeVersion = 0;
	int DriverVersion = 0;
	int NumDevices = 0;

	// --------------------------------------------------
	// ------ GET CUDA RUNTIME AND DRIVER VERSION -------
	cudaRuntimeGetVersion(&RuntimeVersion);
	HANDLE_CUDA(cudaGetLastError());
	cudaDriverGetVersion(&DriverVersion);
	HANDLE_CUDA(cudaGetLastError());

	// ------------------------------------
	// ------- CHECK DRIVER SUPPORT -------
	if (RuntimeVersion > DriverVersion) {
		std::string Message = std::string("Cuda Runtime Version : ") + std::to_string(RuntimeVersion) + std::string("\n");
		Message +=            std::string("Cuda Driver Version  : ") + std::to_string(DriverVersion) + std::string("\n");
		Message += std::string("Internal Error. Insufficient Driver, please update Graphics Drivers to the latest version\n");
		mexErrMsgTxt(Message.c_str());
	}

	// -------------------------------------------
	// ------- CHECK FOR AMOUNT OF DEVICES -------
	cudaError_t Err = cudaGetDeviceCount(&NumDevices);
	if (Err == cudaErrorNoDevice) {
		std::string Message = std::string("Internal Error. No CUDA Capable Device Found!\n");
		mexErrMsgTxt(Message.c_str());
	}

	// -------------------------------------
	// ------- GET DEVICE PARAMETERS -------
	cudaDeviceProp DevProperties;
	cudaStream_t Stream;
	HANDLE_CUDA(cudaGetDeviceProperties(&DevProperties, 0));
	HANDLE_CUDA(cudaStreamCreate(&Stream));
#ifdef DEBUG_PRINT
	mexPrintf("Found CUDA Capable GPU: %s \n",DevProperties.name);
#endif

	int L0_Size = (Nx * Ny) / 1024 + 1;
	int L1_Size = (Nx * Ny) / (1024 * 1024) + 1;
	int L2_Size = (Nx * Ny) / (1024 * 1024 * 1024) + 1; 


	// -------------------------------------
	// ------- ALLOCATE DEVICE MEMORY ------
	size_t ArraySize = Nx * Ny * sizeof(float);  
	size_t RequiredMem = ArraySize * 4;
	size_t FreeMemory = 0;
	size_t TotalMemory = 0;
	HANDLE_CUDA(cudaMemGetInfo(&FreeMemory, &TotalMemory));
	if(RequiredMem > FreeMemory){
		std::string Message = std::string(" Error Not enough graphic memory. Required: ") + std::to_string(RequiredMem / (1024 * 1024)) + std::string(" MB | Available: ") + std::to_string(FreeMemory / (1024 * 1024)) + std::string(" MB\n");
		mexErrMsgTxt(Message.c_str());
	}

	HANDLE_CUDA(cudaMalloc((void**)&Dev_RAir, ArraySize));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_VAir_0, ArraySize));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_VAir_1, ArraySize));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_Mask, ArraySize));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_Max_0,L0_Size * sizeof(float)));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_Max_1,L1_Size * sizeof(float)));
	HANDLE_CUDA(cudaMalloc((void**)&Dev_Max_2,L2_Size * sizeof(float)));
	HANDLE_CUDA(cudaMemset(Dev_Max_0, 0, L0_Size * sizeof(float)));
	HANDLE_CUDA(cudaMemset(Dev_Max_1, 0, L1_Size * sizeof(float)));
	HANDLE_CUDA(cudaMemset(Dev_Max_2, 0, L2_Size * sizeof(float)));

	// ----------------------------------
	// ------ ALLOCATE HOST MEMORY ------
	HANDLE_CUDA(cudaHostAlloc((void**)&VAir, ArraySize, cudaHostAllocDefault));
	HANDLE_CUDA(cudaHostAlloc((void**)&RAir, ArraySize, cudaHostAllocDefault));
	HANDLE_CUDA(cudaHostAlloc((void**)&Mask, ArraySize, cudaHostAllocDefault));

	// -----------------------------------------------
	// ------- SPECIFY KERNEL LAUNCHING PARAMS -------
	dim3 THREADS = dim3(1024, 1, 1);
	dim3 BLOCKS  = dim3(Nx / 1024 + 1, Ny, 1);

	// -------------------------------------------
	// ------ TRANSFORM TO ROW-MAJOR ORDER -------
	// TODO: (Optional: Can be done on GPU)
	ColToRow(In_VAir, VAir, Nx, Ny);
	ColToRow(In_RAir, RAir, Nx, Ny);
	ColToRow(In_Mask, Mask, Nx, Ny);

	// ----------------------------
	// ------ COPY TO DEVICE ------
	HANDLE_CUDA(cudaMemcpy(Dev_RAir,	RAir, ArraySize, cudaMemcpyHostToDevice));
	HANDLE_CUDA(cudaMemcpy(Dev_VAir_0,	VAir, ArraySize, cudaMemcpyHostToDevice));
	HANDLE_CUDA(cudaMemcpy(Dev_VAir_1,	VAir, ArraySize, cudaMemcpyHostToDevice));
	HANDLE_CUDA(cudaMemcpy(Dev_Mask,	Mask, ArraySize, cudaMemcpyHostToDevice));

	int NumIterations = 0;
	bool Convergent = false;
	float CurrentThresh = 13;
	size_t SharedMem = 0;

	// Dynamic Shared Memory for CUDA Kernels
	size_t SharedReduce = sizeof(float) * THREADS.x;

	// ---------------------------------------------------------------------------------------------
	// ----- RUN UNTIL THE CONVERGENCE IS REACHED AND OR MAXIMUM AMOUNT OF ITERATIONS DEPLETED -----
	// ----- NOTE: COMPUTATION MUST BE SYNCHRONIZED. THIS IS DONE VIA DUAL-MEMORY BUFFER
	for(int i = 0;i < MaxIter;i++){
		NumIterations = i;
		// ----------------------------------------------------
		if(SwitchDir){ // ------ Include Memory Barrier -------
		// ----------------------------------------------------

			// LAUNCH KERNEL IN DEFAULT STREAM WITH ZERO SHARED MEMORY
			CalculatePotencials<<< BLOCKS, THREADS, SharedMem, Stream >>>(Dev_RAir, Dev_VAir_0, Dev_VAir_1,Dev_Mask, Nx, Ny, Omega, Grid8);
			HANDLE_CUDA(cudaGetLastError());

			// OVERWRITE THE ORIGINAL BUFFER
			HANDLE_CUDA(cudaMemcpyAsync(Dev_VAir_0, Dev_VAir_1, sizeof(float) * Nx * Ny, cudaMemcpyDeviceToDevice,Stream));
				

		// ---------------------------------------------------------------
		}else{ // ------- Do not Use MemBarrier, Update on the Fly -------
		// ---------------------------------------------------------------

			// LAUNCH KERNEL IN DEFAULT STREAM WITH ZERO SHARED MEMORY
			CalculatePotencials<<< BLOCKS, THREADS, SharedMem, Stream >>>(Dev_RAir, Dev_VAir_0, Dev_VAir_0,Dev_Mask, Nx, Ny, Omega, Grid8 );
			HANDLE_CUDA(cudaGetLastError());
		}

		
		// -------------------------------------------------------------------
		// MANUAL 3-STAGE MAXIMUM PARALEL REDUCE. Max Elements: 1024*1024*1024
		// NOTE: This could be improved by using Cublas
		ReduceMax <<<BLOCKS, THREADS, SharedReduce, Stream >>> (Dev_RAir,Dev_Max_0,Nx*Ny);
		HANDLE_CUDA(cudaGetLastError());
		ReduceMax <<<BLOCKS, THREADS, SharedReduce, Stream>>> (Dev_Max_0,Dev_Max_1,L0_Size);
		HANDLE_CUDA(cudaGetLastError());
		ReduceMax <<<BLOCKS, THREADS, SharedReduce, Stream>>> (Dev_Max_1,Dev_Max_2,L1_Size);
		HANDLE_CUDA(cudaGetLastError());
		HANDLE_CUDA(cudaMemcpyAsync(&CurrentThresh, Dev_Max_2 , sizeof(float), cudaMemcpyDeviceToHost,Stream));

		// -----------------------------------------------------------------------------------
		// AT THE END OF EACH CYCLE ,SYNCHRONIZE THE STREAM (Could as well use default Stream)
		HANDLE_CUDA(cudaStreamSynchronize(Stream));


		if(CurrentThresh < Thresh){
			mexPrintf("Convergence after %d Interations with Threshold: %8.7f\n",i, CurrentThresh);
			Convergent = true;
			break;
		}

	}

	if(!Convergent){
		std::string Message = std::string("Computation Convergence not reached with specified Threshold and amount of Iterations!\n");
		Message += std::string("Amount of Iterations: ") + std::to_string(MaxIter) + std::string(" | Threshold: ") + std::to_string(Thresh) + std::string("\n");
		Message += std::string("Final Threshold: ") + std::to_string(CurrentThresh) + std::string("\n");
		mexWarnMsgTxt(Message.c_str());
	}

	// -------------------------------
	// ------ COPY BACK TO HOST ------
	HANDLE_CUDA(cudaMemcpy(VAir, Dev_VAir_0, ArraySize, cudaMemcpyDeviceToHost));
	HANDLE_CUDA(cudaMemcpy(RAir, Dev_RAir, ArraySize, cudaMemcpyDeviceToHost));
	
	// ---------------------------------------------
	// ------ ASSOCIATE OUTPUT AS A 2D Matrix ------
	const size_t OutputSize[] = { static_cast<size_t>(Ny) ,static_cast<size_t>(Nx) };
	const size_t SingleSize[] = { static_cast<size_t>(1), static_cast<size_t>(1) };
	plhs[0] = mxCreateNumericArray(2, OutputSize, mxSINGLE_CLASS, mxREAL);
	plhs[1] = mxCreateNumericArray(2, OutputSize, mxSINGLE_CLASS, mxREAL);
	plhs[2] = mxCreateNumericArray(2, SingleSize, mxINT32_CLASS,  mxREAL);

	float * Out_Data_VAir = static_cast<float*>(mxGetData(plhs[0]));
	float * Out_Data_RAir = static_cast<float*>(mxGetData(plhs[1]));
	int * Out_Data_Iter = static_cast<int*>(mxGetData(plhs[2]));
	Out_Data_Iter[0] = NumIterations;

	// --------------------------------------------------
	// ------ TRANSFORM BACK TO COLUMN MAJOR ORDER ------
	// TODO: (Optional): Could as well be done on GPU
	RowToCol(Out_Data_VAir,VAir, Nx, Ny);
	RowToCol(Out_Data_RAir,RAir, Nx, Ny);

	// ------------------------------------
	// ------ CLEAN UP DEVICE MEMORY ------
	if(Dev_RAir   != nullptr){  HANDLE_CUDA(cudaFree(Dev_RAir)); Dev_RAir = nullptr; }
	if(Dev_VAir_0 != nullptr){  HANDLE_CUDA(cudaFree(Dev_VAir_0));Dev_VAir_0 = nullptr; }
	if(Dev_VAir_1 != nullptr){  HANDLE_CUDA(cudaFree(Dev_VAir_1)); Dev_VAir_1 = nullptr; }
	if(Dev_Mask   != nullptr){  HANDLE_CUDA(cudaFree(Dev_Mask));Dev_Mask = nullptr; }
	if(Dev_Max_0  != nullptr){  HANDLE_CUDA(cudaFree(Dev_Max_0));Dev_Max_0 = nullptr; }
	if(Dev_Max_1  != nullptr){  HANDLE_CUDA(cudaFree(Dev_Max_1));Dev_Max_1 = nullptr; }
	if(Dev_Max_2  != nullptr){  HANDLE_CUDA(cudaFree(Dev_Max_2));Dev_Max_2 = nullptr; }
	HANDLE_CUDA(cudaStreamDestroy(Stream));

	// -----------------------------------
	// ------ CLEAN UP HOST MEMORY -------
	if(VAir != nullptr){ HANDLE_CUDA(cudaFreeHost(VAir));VAir = nullptr;}
	if(RAir != nullptr){ HANDLE_CUDA(cudaFreeHost(RAir));RAir = nullptr;}
	if(Mask != nullptr){ HANDLE_CUDA(cudaFreeHost(Mask));Mask = nullptr;}

	return;
}