%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% @brief Calculate Electric Field Distribution in the Area
% @VOrig = Original voltage Field
% @StripInd = Matrix with 0 & 1 (1 = Posittion of a stripline) "IE: Mask"
% @NH: Discretization Grid Height (Y - Dimension)
% @NL: Discretization Grid Width  (X - Dimension)
%@Grid8 : Uses either 4 or 8 points for stepsize
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [VAirOut,Iter] = CalculatePotencials( VOrig,StripInd,NH,NL,Grid8)

global S;
S.RAir = single(zeros(NH,NL)); % Init
S.VAir = single(VOrig);        % Init
VAirOut = single(VOrig);       % Init


if(S.UseCPU)
    tic;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % LOOP FOR MAXIMUM AMOUNT OF OTERATIONS 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    ConvergenceReached = 0;
    for Iter = 1:S.MaxIter
        disp(['Processing: ',num2str(Iter),' / ',num2str(S.MaxIter)]);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        for y = 2:NH - 1 % Ommit Border Points
            for x = 2:NL - 1 % Ommit Border Points

                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                % Continue Updating matrix only if we are not updating microstrip values%
                if(StripInd(y,x) ~= 1)
                    

                    % EXTRACT NEARBY CELLS 
                    VT  = S.VAir(y-1,x); 
                    VB  = S.VAir(y+1,x); 
                    VL  = S.VAir(y,x-1); 
                    VR  = S.VAir(y,x+1); 
                    WC  = S.VAir(y,x);   
                    VUR = S.VAir(y-1,x+1);
                    VUL = S.VAir(y-1,x-1);
                    VBR = S.VAir(y+1,x+1);
                    VBL = S.VAir(y+1,x-1);


                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Calculate Air regardless of Permitivity
                    C0 = 1/sqrt(2);
                    C1 = 1 - C0;
                    Rep_0 = 1.0/4.0* (VT + VB + VL + VR);     % Perpendicular
                    Rep_1 = 1.0/4.0* (VUR + VUL + VBR + VBL); % Diagonal Elements
                    Rep8 = C0*Rep_0 + C1 * Rep_1 - WC; 
                    Rep4 = Rep_0 - WC;
                    
                    
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Choose Computation Type %
                    if(Grid8 == 1)
                        S.RAir(y,x) = Rep8;
                    else
                        S.RAir(y,x) = Rep4;
                    end
                    
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Note: If Mem Sync is enabled, then the entire field is updated at once. "Value at [1,2] is not influenced by [1,1]"
                    % If Mem Sync is disabled, then values are updated on-the-fly. "Value at [1,2] is influenced by previous result - [1,1]"
                    if(S.MemSync)
                        VAirOut(y,x) = WC + S.Omega*S.RAir(y,x);
                    else
                        S.VAir(y,x) = WC + S.Omega*S.RAir(y,x);
                    end
                    

                end

            end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        if(S.MemSync)
            S.VAir = VAirOut;
        end


        % CHECK CONVERGENCE
        Rmax = max(S.RAir(:));
        if(Rmax < S.Thresh)
            ConvergenceReached = 1;
            disp(['Done After ',num2str(Iter),' Iterations |  Thresh: ',num2str(Rmax)]);
            break;
        end
    end

    % Display if the Algorithm has not Converged
    if(ConvergenceReached == 0)
        warning('Warning: Not enough Iterations and or high Threshold Value!');
    end

    CPU_Stop = toc;
    disp(['CPU Elapsed Time: ',num2str(CPU_Stop),' Sec']);

else
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% USE CUDA ACCELERATION %%%%%%%
    % TODO: (Optional) Under heavy repetetion calls,the library can be updated so that all resources remain allocated across calls
    % This would overall increase the performance at an extra cost of having to manually deallocate resources after the lib is no longer used.
    tic;
    Params = single([S.Omega,S.MaxIter,S.Thresh,S.MemSync,Grid8]);
    [S.VAir,S.RAir,Iter] = CudaFEM(S.VAir,S.RAir,int32(StripInd),Params);
    GPU_Stop = toc;
    disp(['GPU Elapsed Time: ',num2str(GPU_Stop),' Sec']);
    %%%%%%% USE CUDA ACCELERATION %%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end

%%%%%%%%%%%%%%%%%%%%
% ASSIGN TO OUTPUT %
VAirOut = S.VAir;


end

