close all;
clear all;
clc;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% @brief: Laplace Equation Simulation in Matlab and Accelerated by CUDA 
% IrisCores.com vojta.ters@iriscores.com
% The Shielding is Expected to be at 0V as well as one of the Microstrip Lines.
% Microstrips should be close to each other and away from the shielding in order that the Impedance is calculated between
% them and not between the shielding. Impedance of the stripline is not estimated within this demo. Its purpose is to show 
% The possible acceleration with CUDA.
% It would be required to calculate curvature integral from the definition of Divergence Theorem.
% NOTE: Change of V0 has insignificant impact on the Impedance of the Microstrip Line.
% NOTE: Amount of Points (Grid Discretization) influence the resulting Impedance.
% NOTE: --> This demo doesnt use Relative Permitivity - It Expects a constant permitivity across the entire region.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add release Directory to Matlab Path %
MexFilePath = ['..',filesep,'x64',filesep,'Release'];
addpath(MexFilePath);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% Get Display Resolution %%%%%%%
global S;
S.Displ = groot();
S.ScW = S.Displ.ScreenSize(3);
S.ScH = S.Displ.ScreenSize(4);
S.FiW = 1024;
S.FiH = 720;
S.Fs  = 17;
S.ContourLines = 17;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Display Border Limitations %
Ma            = 0.08;
Left          = Ma;
Bot           = Ma;

S.V0 = 1;                % Stripline voltage [V]
S.Vgnd = 0;              % Ground Voltage = 0V

S.d = 0.6;               % Distance between Striplines in [mm]
S.W = 1.1;               % Microstrip Width in [mm]
S.h = 4;                 % Thickness of the Substrates in [mm]
S.Thresh = 1e-4;         % Threshold for convergence error
S.MaxIter = 1000;        % Maximum Amount of Iterations
S.Omega = 1.0;           % "SuperRelaxation Coefficient" [1 - 1.4]

S.S = 12*S.d;             % Width of the Upper StripLine in mm
S.L = 20*S.d;             % X Dimension in mm
S.H = S.d + 2*S.h;        % Y Dimension in mm
S.PointsPerMm = 20;       % Amount of Points per mm
S.UseCPU = 1;             % 1 --> CPU | 0--> CUDA
S.MemSync = 1;            % Memory Barrier Synchronization


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Integer dimensions of DUT Area %
NL = round(S.PointsPerMm*S.L); % Total Grid Width
NH = round(S.PointsPerMm*S.H); % Total Grid Heigh
Nd = round(S.PointsPerMm*S.d); % Distance between Striplines in Points
Nh = round(S.PointsPerMm*S.h); % Substrate Height in Points
Nw = round(S.PointsPerMm*S.W); % Stripline Width in Points



VOriginal = zeros(NH,NL);
StripInd  = zeros(NH,NL);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% SET INITIAL VOLTAGE %%%%%%%
%LEFT AND RIGHT SIDEs
for ii = 1:NH
    VOriginal(ii,1) = S.Vgnd;
    VOriginal(ii,NL) = S.Vgnd;
end
%TOP AND BOTTOM SIDEs
for jj = 1:NL
    VOriginal(1,jj) = S.Vgnd;
    VOriginal(NH,jj) = S.Vgnd;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ADD MICROSTRIPS AND THEIR INDICATORS
% NOTE: Lower Microstrip has voltage
stripBegin = round(NL/2 - Nw/2);
for jj = stripBegin:stripBegin+Nw
    VOriginal(Nh + 1,jj)      = S.Vgnd;
    VOriginal(Nd + Nh - 1,jj) = S.V0;
    StripInd(Nh + 1,jj) = 1;
    StripInd(Nd + Nh-1,jj) = 1;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% SET AVERAGE VOLTAGE ACROSS ALL NODES (INITIAL GUESS)%%%
% NOTE: Set to (V0 + V1) / 2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for ii = 2:NH - 1
    for jj = 2:NL - 1
        if(StripInd(ii,jj) ~= 1)
            VOriginal(ii,jj) = (S.Vgnd + S.V0)/2.0;
        end
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% COMPUTE THE DIFFERENTIAL LAPLACE EQUATION %%%%%%%
[Air4,Iter4] = CalculatePotencials(VOriginal,StripInd,NH,NL,0);
[Air8,Iter8] = CalculatePotencials(VOriginal,StripInd,NH,NL,1);
 disp(['Convergence4: ',num2str(Iter4),' Iterations']);
 disp(['Convergence8: ',num2str(Iter8),' Iterations']);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% CHECK RANGES OF TICKMARKS %%%%%%%
Min4 = min(Air4(:));
Max4 = max(Air4(:));
Min8 = min(Air8(:));
Max8 = max(Air8(:));

if(isinf(Min4) || isnan(Min4))
    Min4 = 0;
end
if(isinf(Max4) || isnan(Max4))
    Max4 = 1;
end

if(isinf(Min8) || isnan(Min8))
    Min8 = 0;
end

if(isinf(Max8) || isnan(Max8))
    Max8 = 1;
end

Marks4 = linspace(Min4,Max4,S.ContourLines);
Marks8 = linspace(Min8,Max8,S.ContourLines);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% DISPLAY ELECTRIC FIELD DISTRIBUTION  %%%%%%%
F0 = figure ('name','Electric Field Distribution - 4 Samples','position',[S.ScW/2 - S.FiW/2,S.ScH/2 - S.FiH/2,S.FiW,S.FiH]);
contour(Air4,S.ContourLines,'LineWidth',2);
hcb = colorbar;
set(gca, 'XTick', []);
set(gca, 'YTick', []);
set(hcb, 'YTick',Marks4);
title('AIR 4 - Grid','FontSize',S.Fs);
colormap morgenstemning;
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);

F1 = figure('name','Electric Field Distribution - 8 Samples','position',[S.ScW/2 - S.FiW/2,S.ScH/2 - S.FiH/2,S.FiW,S.FiH]);
contour(Air8,S.ContourLines,'LineWidth',2);
hcb = colorbar;
set(gca, 'XTick', []);
set(gca, 'YTick', []);
set(hcb, 'YTick',Marks8);
title('AIR 8 - Grid','FontSize',S.Fs);
colormap morgenstemning;
set(gca,'Position',[Left Bot 1 - 2*Ma 1 - 2*Ma]);


%%%%%%%%%%%%%%%%%%
% Clean Mex File %
clear mex;




